#!/usr/bin/perl
#
# clipinc.pl
#
# A script that implements stream tag integration for the video disk rekorder
# (VDR) by Klaus Schmidinger.
#
# This script is licensed under the GPL (Version 2).
# (C)Copyright 2004 by Robert Huitl
#
# Parts of the code are derived from a script called getcidata.pl and the
# vdr-sharemarks project.
#
#
# IMPORTANT: Change the first few variables in vdrrecinfo.pl to match your
# setup! Do the same for this script (clipinc.pl).
#
# IMPORTANT: You need to start VDR with an additional parameter:
#   -r /etc/vdr/scripts/clipinc/vdrrecinfo.pl
# This script will store important information about a recording, such as
# information about the channel and the recording time in a file called
# "recinfo.conf" in the recording directory.
# 
# IMPORTANT: You will need a license key for ClipInc. in order to use this
# script. There are two ways to get one:
# 1. Install and start the Windows client, open the preferences, go to the
#    server-tab and take the first 11 characters from the "license key" text box
# 2. Open http://www.geschickter-als-kaufen.de/hilfe.htm and enter your email
#    address or cell-phone number.
# Set $LICENSE below to this number in the format "xxxxx-xxxxx".
#
# IMPORTANT: If you get an error like this:
#   "clipinc.pl: Cannot map stream-id XXXX to a ClipInc-ID"
# you need to edit clipinc.map and map the stream-id to the corresponding
# Clip-Inc-ID. You can get a list of senders offered by the ClipInc. server by
# running:
#   ./clipinc.pl senders
# The first number on each line is the ClipInc-ID.

use strict;
use warnings;
use LWP::UserAgent;
use Time::Local;
use File::Path;
require "clipinc.map";

my $CLIPS_DIR = "/video/Clips";	# Where to store the clips (make sure this
				#   directory exists!)
my $REC_FPS = 25;		# NTSC users change this to 30, needed to
				#   calculate recording length and to map a time
				#   offset in the recording to the frame number
my $LICENSE = "xxxxx-xxxxx";	# Put the first 11 digits of your license here
				#   Get your license for free at:
				#   http://www.geschickter-als-kaufen.de/hilfe.htm
				#   (scroll down to "Anfordern einer Lizenznummer")
my $TOUCH_UPDATE_FILE = "/video/.update";	# 'touch' update file whenever
						# recordings are created (VDR
						# 1.3.11+ needs this); empty
						# string to disable
my $VDRCUT = "./vdrcut";

##
## no configuration changes should be required below this line
##
my $NAME = "clipinc.pl";
my $VERSION = "0.1dev";
my $CLIPINC_HOST = "community.clipinc.com";
my $CLIPINC_VERSION = 122;

my $recording = shift(@ARGV);
if(!$recording) {
	PrintUsage();
	exit;
}
if($recording eq 'senders') {
	PrintSenderList();
	exit;
}
my($r_path,$r_beginning,$r_end,$r_ci_id) = GetRecordingInfo($recording);
if(!$r_path) {		# GetRecordingInfo() failed
	exit;
}
my($offset,@stags) = GetStreamTags($r_ci_id, $r_beginning, $r_end);
if(defined($offset)) {	# GetStreamTags() succeeded
	ExtractStreams($r_path, $r_beginning, $r_end, $offset, @stags);
}
exit;
# End of "main" :-)


sub PrintUsage
{
	print qq^$NAME $VERSION

Syntax: $0 [recording-directory | "senders"]

This script extracts video clips from the given recording into
seperate files named after the song title. This is accomplished by
downloading so-called stream tags from the free Tobit ClipInc. service.

You can get a list of all ClipInc. senders by starting this script with
the parameter "senders" instead of a recording directory.
^;
}

sub GetRecordingInfo
{
	my $rec_dir = shift;
	$rec_dir .= '/' if($rec_dir !~ /\/$/);
	my $recinfo_fn = $rec_dir . 'recinfo.conf';
	my $index_fn = $rec_dir . 'index.vdr';
	my $index_size = -s $index_fn;
	if(!$index_size) {
		print "$NAME: Cannot open $index_fn, aborting\n";
		return;
	}
	my $rec_frames = $index_size / 8;	# One index per frame,
						# 8 bytes per index
	my $rec_seconds = $rec_frames / $REC_FPS;
	
	my($sid,$rstart) = ReadRecordingInfo($recinfo_fn);
	if(!$sid) {
		print "$NAME: Cannot open $recinfo_fn\n";
		return;
	}
	my $ci_id = $ClipIncMap::CLIPINC_MAP{$sid};
	if(!$ci_id) {
		print "$NAME: Cannot map stream-id $sid to a ClipInc-ID\n";
		return;
	}

	my $rec_path = GetAbsolutePath($rec_dir);
	my $rec_beginning = $rstart;
	my $rec_end = $rec_beginning + $rec_seconds;

	print "Info about $rec_path:\n";
	print "Rec-Start: ".TimeUnixToClipInc($rstart)."\n";
	print "End:       ".TimeUnixToClipInc($rec_end)."\n";
	print "Duration:  $rec_seconds\n";
	print "Stream-ID: $sid (ClipInc-ID: $ci_id)\n\n";
	return($rec_path,$rec_beginning,$rec_end,$ci_id);
}

sub GetAbsolutePath
{
	my $path = shift;
	if($path =~ /^\//) {
		# Absolute path
		return $path;
	} else {
		# Relative path
		chomp(my $cwd = `pwd`);
		return $cwd.'/'.$path;
	}
}

sub ReadRecordingInfo
{
	my $fn = shift;
	if(!open(CONF, "<$fn")) {
		return;
	}
	my($channel,$nchid,$beginning,$end,$rstart);
	foreach(<CONF>) {
		if($_ =~ /channel\s+(.*)/) {
			$channel = $1;
		} elsif($_ =~ /chid\s+(.*)/) {
			$nchid = $1;
		} elsif($_ =~ /^start\s+(.*)/) {
			$beginning = $1;
		} elsif($_ =~ /end\s+(.*)/) {
			$end = $1;
		} elsif($_ =~ /rstart\s+(.*)/) {
			$rstart = $1;
		}
	}
	close(CONF);
	my($sid) = $channel =~ /-(\d+)$/;
	#return($sid,$channel,$nchid,$beginning,$end,$rstart);
	return($sid,$rstart);
}

sub GetStreamTags
{
	# Attempts to download stream tags for a certain channel
	# and a certain timespan from the ClipInc. server
	#
	# Arguments: 	ClipInc-ChannelID
	#		Date of oldest stream tag to fetch (unix format)
	#		Date of newest stream tag to return (unix format)
	my $channelid = shift;
	my $lastdate = TimeUnixToClipInc(shift);
	my $newestdate = shift;
	my $ua = LWP::UserAgent->new(
		agent	=> "Tobit ClipInc"
	);

	my $response = $ua->get(
		"http://" . $CLIPINC_HOST . "/getStreamTags.asp".
		"?ChannelID="	.	$channelid.
		"&LastDate="	.	$lastdate.
		"&Version="	.	$CLIPINC_VERSION.
#		"&Channel="	.	,.
		"&Site-ID="	.	$LICENSE.
		"&Time-Ref="	.	TimeUnixToClipInc()
#		"&Time-Ref=25.03.2004 23:35:00.00"
	);
	if(!$response->is_success()) {
		print "$NAME: Cannot download stream tags for channel $channelid: ";
		print $response->status_line, "\n";
		return;
	}

	# Split lines at <br> and parse return data
	my($st_count,$st_vintern,$st_timeref,$st_offset);
	my @streamtags;
	my @lines = split(/<br>/, $response->content);
#print ">> ", $response->content, " <<\n";
	if(scalar(@lines) == 0) {
		print "$NAME: Server reply is empty",
			$LICENSE eq 'xxxxx-xxxxx'
				? " - please set \$LICENSE and read the README!\n"
				: " - invalid license?\n";
		return;
	}

	foreach(@lines) {
		if($_ =~ /^count=(\d*)/) {
			$st_count = $1;
		} elsif($_ =~ /^vintern=(\d*)/) {
			$st_vintern = $1;
		} elsif($_ =~ /^time-ref=(.*)/) {
			$st_timeref = $1;
		} elsif($_ =~ /^offset=(.*)/) {
			$st_offset = $1;
 		} else {
			$_ =~ /(.*?)\s(.*?\s.*?)\s(.*?\s.*?)\s"(.*?)"\s(\d+)\s(\d+)\s"(.*?)"\s"(.*?)"\s"(.*?)"/;
			my %streamtag = (
				StationID	=> $1,
				StartDate	=> $2,
				EndDate		=> $3,
				Title		=> $4,
				Genre		=> $5,
				Whatever1	=> $6,
				Interpret	=> $7,
				Album		=> $8,
				Web		=> $9
			);

			if(TimeClipIncToUnix($streamtag{EndDate}) < $newestdate) {
				push @streamtags, \%streamtag;
			}
		}
	}

	print "Information about stream tags:\n";
	print "Count:    $st_count (".scalar(@streamtags)." after filtering)\n";
	print "VIntern:  $st_vintern\n";
	print "Time-Ref: $st_timeref\n";
	print "Offset:   $st_offset\n\n";

	return $st_offset, @streamtags;
}

sub ExtractStreams
{
	# Extracts one or more streams identified by stream tags from the
	# recording. $offset is used to correct differing times on this machine
	# and the ClipInc. server.
	
	my($r_path,$r_beginning,$r_end,$offset,@stags) = @_;
	my $r_offset = CalculateOffset($offset);	# Correct time skews
	print "Offset[s]: $r_offset\n";
	
	my $nExtracted = 0;				# Counter variable
	foreach(@stags) {
		print "\n$_->{StartDate} $_->{EndDate} $_->{Interpret} - $_->{Title}";
		print " (from $_->{Album})" if($_->{Album} ne '');
		# Get start and end date from stream tag and apply the offset
		my $startdate = TimeClipIncToUnix($_->{StartDate});
		my $enddate = TimeClipIncToUnix($_->{EndDate});
		my $offset_begin = $startdate - $r_beginning - $r_offset;
		my $offset_end = $enddate - $r_beginning - $r_offset;
		
		print " | offset: $offset_begin - $offset_end";
		print " [Start not recorded]" if($offset_begin < 0);
		print " [End not recorded]" if($enddate > $r_end);	# TODO: Bug? no offset applied?!
		print "\n";
		
		my $startframe = $offset_begin < 0 ? 0 : ($offset_begin*$REC_FPS);
		my $endframe = $offset_end*$REC_FPS;
		
		if($endframe > 0) {
			# Create recording in appropriate directory
			my($rec_dir) = $r_path =~ /.*\/(.*?\.rec)\//;
			my $dest_dir = $CLIPS_DIR.
				'/'.EscapeFilename($_->{'Interpret'}.' - '.$_->{'Title'}).
				'/'.$rec_dir;
			mkpath($dest_dir);
			
			if(system("$VDRCUT $r_path \"$dest_dir/001.vdr\" $startframe $endframe") != 0) {
				print "$NAME: execution of $VDRCUT failed\n";
			} else {
				if(!rename("$dest_dir/001.vdr.index", "$dest_dir/index.vdr")) {
					print "$NAME: Warning: cannot rename index file $dest_dir/001.vdr.index to index.vdr!\n";
				}
				$nExtracted++;
			}
		} else {
			print "skipping (ends before start of recording)\n";
		}
	}
	if(($nExtracted > 0) && ($TOUCH_UPDATE_FILE ne '')) {
		# (Some) clips were extracted and update file is to be touched
		# so VDR 1.3.11+ will re-read the recordings directory tree
		`touch $TOUCH_UPDATE_FILE`
	}
}

sub EscapeFilename
{
	# TODO: Improve filename mangling
	my $fn = shift;
	$fn =~ s/ /_/g;		# Replace whitespaces
	return $fn;
}

sub CalculateOffset
{
	my $offset = shift;
	# The ClipInc. server returns offset in format "01.01.1900..."
	# but unix time functions seem to have a problem with
	# dates before 1970 so we simply add 100 (which makes no
	# difference when calculating an offset)
	$offset =~ /^(.*)(\d{4})(.*)$/;
	$offset = TimeClipIncToUnix($1.($2+100).$3);

	# The offset is expressed in relation to 01.01.1900 00:00:00.000,
	# but keep in mind that we added +100 years
	my $nulldate = TimeClipIncToUnix('01.01.2000 00:00:00.000');
	
	return $offset - $nulldate;
}

sub TimeUnixToClipInc
{
	my $value = shift || time;
	my($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) =
		localtime($value);
	return sprintf('%02d.%02d.%04d %02d:%02d:%02d.%03d',
		$mday, $mon+1, $year+1900,
		$hour, $min, $sec, 0);
	# Please note that we set msecs to 0
}

sub TimeClipIncToUnix
{
	my $value = shift;
	my($day,$month,$year,$hour,$min,$sec,$nsec)
		= $value =~ /(\d{2}).(\d{2}).(\d{4}) (\d{2}):(\d{2}):(\d{2})\.(\d{3})/;
	return timelocal($sec,$min,$hour,$day,$month-1,$year-1900);
	# Please note that we dropped nsec.
}

sub PrintSenderList
{
	# Attempts to download a list of all senders
	my $ua = LWP::UserAgent->new(
		agent	=> "Tobit ClipInc"
	);

	my $response = $ua->get(
		"http://" . $CLIPINC_HOST . "/channellist.asp".
		"?version="	.	$CLIPINC_VERSION
#		"&Site-ID="	.	$LICENSE
	);
	if(!$response->is_success()) {
		print "$NAME: Cannot download sender list: ";
		print $response->status_line, "\n";
		return;
	}
	# Split output at <br> tokens
 	my @lines = split(/<br>/, $response->content);
	foreach(@lines) {
		print $_, "\n";
	}
	print "Senders in list: ", scalar(@lines), ". The number at the beginning is the ClipInc.-ID.\n";
}

#sub tm2unix					# borrowed from marks2pts
#{
#	my $value = shift || time;
#	my $calc;
#
#	if($value =~ /^\d+$/) {
#		my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) =
#		localtime($value);
#		$calc = sprintf('%02d.%02d.%04d %02d:%02d',
#		$mday, $mon+1, $year + 1900,
#		$hour, $min);
#	} elsif ($value =~ /(\d{2})\.(\d{2})\.(\d{4})\s+(\d{2})\:(\d{2})\:(\d{2})/) {
#		$calc = timelocal($6, $5, $4, $1, $2-1, $3);
#	} elsif($value =~ /(\d{2})\.(\d{2})\.(\d{4})\s+(\d{2})\:(\d{2})/) {
#		$calc = timelocal(0, $5, $4, $1, $2-1, $3);
#	} else {
#		return error('Wrong Format: %s!', $value);
#	}
#	return $calc;
#}
#
#sub TimeVDRToUnix
#{
#	my $value = shift;
#	my($year,$month,$day,$hour,$min,$sec,$msec)
#		= $value =~ /(\d{4})-(\d{2})-(\d{2})\.(\d{2})[:\.](\d{2})\.(\d{2})\.(\d{2})/;
#	return timelocal($sec,$min,$hour,$day,$month-1,$year-1900);
#	# Please note that we dropped msec.
#}
#
