/*
 * vdrcut.c	extracts a specified timespan from a vdr recording.
 *		parts are based on genindex.c.
 *
 * Compile:	gcc -o vdrcut -O2 -Wall -funroll-loops vdrcut.c
 *
 * Authors:	Robert Huitl
 * License:	GPL Version 2
 *
 */
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>

#define NAME "VDRCut"
#define VERSION "0.0.1dev"
#define INDEXFILE "index.vdr"

#define NO_PICTURE	0
#define I_FRAME		1
#define P_FRAME		2
#define B_FRAME		3

/* VDR supports only offset with sizeof(int) */
struct tIndex {int offset; unsigned char type; unsigned char number; short reserved; };
typedef struct tIndex TIndex;

static FILE *fp, *fp2, *fp3;

int ExtractFrames(char **argv, TIndex *idxStart, TIndex *idxEnd);
int ExtractIndex(char **argv, int nStart, int nEnd);

int main(int argc, char **argv)
{
	if(argc <= 4) {
		printf("%s %s\n\n", NAME, VERSION);
		printf("Syntax: %s recording-directory output-file first-frame last-frame\n\n", argv[0]);
		return(1);	
	}
	
	int n, nStart, nEnd, findStartI;
	TIndex index, idxStart, idxEnd;//, idxLastI;
	char *indexfile = (char *)malloc(strlen(argv[1]) + strlen(INDEXFILE) + 1);
	sprintf(indexfile, "%s%s", argv[1], INDEXFILE);
	nStart = atoi(argv[3]);
	nEnd = atoi(argv[4]);
	
	printf("Extracting frames %d to %d\n", nStart, nEnd);
	
	fp = fopen(indexfile, "r");
	if(!fp) {
		if (errno != ENOENT) {
			fprintf(stderr, "Could not open %s: %s\n", indexfile, strerror(errno));
			return -1;
		}
		return(1);
	}
	int nFrames = 0;
	findStartI = 0;
	while(1) {
		n = fread(&index, sizeof(index), 1, fp);
		if(n != 1) {
			if(feof(fp))
				break;
			perror("Could not read index file");
		} else {
			/*if(index.type == I_FRAME) {
				idxLastI = index;
				if(findStartI) {
					idxStart = index;
					printf("Start IFrame at offset %d\n", index.offset);
					findStartI = 0;
				}
			}*/
			if(nFrames == nStart) {
				//if(index.type == I_FRAME)
					idxStart = index;
				/*else
					findStartI = 1;*/
			} else if((nFrames == nEnd) || (nFrames == (nEnd+1))) {
				// TODO: Make sure the entire last frame is
				// copied if nEnd points to the last frame in
				// the recording (the hack above only works when
				// nEnd is not the last frame)
				idxEnd = index;
			}
			nFrames++;
		}	
	}	
	fclose(fp);
	
	//printf("Frames:       %d\n", nFrames);
	//printf("Start offset: %d (%03d)\n", idxStart.offset, idxStart.number);
	//printf("End offset:   %d (%03d)\n", idxEnd.offset, idxStart.number);
	
	ExtractFrames(argv, &idxStart, &idxEnd);
	ExtractIndex(argv, nStart, nEnd);
	
	free(indexfile);
	return(0);
}

#define BUFSIZE 8192

int ExtractFrames(char **argv, TIndex *idxStart, TIndex *idxEnd)
{
	char *rec_file = (char *)malloc(strlen(argv[1]) + strlen("001.vdr") + 1);
	int nCurNumber = idxStart->number;
	int nOffset, n;
	char fn[8];
	unsigned char buf[1];	// BUFSIZE
	
	sprintf(fn, "%03d.vdr", nCurNumber);
	sprintf(rec_file, "%s%s", argv[1], fn);

	fp3 = fopen(argv[2], "w");
	if(!fp3) {
		if (errno != ENOENT) {
			fprintf(stderr, "Could not open %s: %s\n", rec_file, strerror(errno));
		}
		free(rec_file);
		return(1);
	}
	fp2 = fopen(rec_file, "r");
	if(!fp2) {
		if (errno != ENOENT) {
			fprintf(stderr, "Could not open %s: %s\n", rec_file, strerror(errno));
		}
		fclose(fp3);
		free(rec_file);
		return(1);
	}	
	//printf("Seeking to %d ", idxStart->offset);
	fseek(fp2, idxStart->offset, SEEK_SET);
	nOffset = ftell(fp2);
	//printf("offset is %d\n", nOffset);
	
	int nBytes = 0;
	// We copy from the first frame's offset
	// to the last+1 frame's offset minus 1
	while((nCurNumber < idxEnd->number) || (nOffset <= (idxEnd->offset-1))) {
		n = fread(buf, 1, 1, fp2);
		if(n != 1) {
			if(feof(fp2)) {
				fclose(fp2);
				nCurNumber++;
				sprintf(fn, "%03d.vdr", nCurNumber);
				sprintf(rec_file, "%s%s", argv[1], fn);
				
				fp2 = fopen(rec_file, "r");
				if(!fp2) {
					break;
				}
				nOffset = 0;
			}
		} else {
			nOffset++;
			nBytes++;
			fwrite(buf, 1, 1, fp3);
		
			if((nBytes % (1024*100)) == 0) {
				printf("\rCopying %d kbytes...", nBytes/1024);
			}
		}
	}
	fclose(fp2);
	fclose(fp3);

	printf("\n");	
	free(rec_file);
	return(0);
}

int ExtractIndex(char **argv, int nStart, int nEnd)
{
	char *idx_file = (char *)malloc(strlen(argv[1]) + strlen("index.vdr") + 1);
	char *idxn_file = (char *)malloc(strlen(argv[2]) + strlen(".index") + 1);
	int n;
	TIndex idxBuf;
	
	sprintf(idx_file, "%s%s", argv[1], "index.vdr");
	sprintf(idxn_file, "%s%s", argv[2], ".index");

	fp3 = fopen(idxn_file, "w");
	if(!fp3) {
		if (errno != ENOENT) {
			fprintf(stderr, "Could not open %s: %s\n", idxn_file, strerror(errno));
		}
		free(idx_file);free(idxn_file);
		return(1);
	}
	fp2 = fopen(idx_file, "r");
	if(!fp2) {
		if (errno != ENOENT) {
			fprintf(stderr, "Could not open %s: %s\n", idx_file, strerror(errno));
		}
		fclose(fp3);
		free(idx_file);free(idxn_file);
		return(1);
	}	
	//printf("Seeking to %d ", idxStart->offset);
	fseek(fp2, nStart*sizeof(idxBuf), SEEK_SET);
	//printf("offset is %d\n", nOffset);
	
	int nFrame = 0;
	int nOffsetStart = 0;
	while(nFrame <= (nEnd - nStart)) {
		n = fread(&idxBuf, sizeof(idxBuf), 1, fp2);
		if(n != 1) {
			if(feof(fp2)) {
				break;
			}
		} else {
			if(nFrame == 0)
				nOffsetStart = idxBuf.offset;
			idxBuf.offset -= nOffsetStart;
			idxBuf.number = 1;			// TODO: No splitting for now
			fwrite(&idxBuf, sizeof(idxBuf), 1, fp3);
			
			if((nFrame % (100)) == 0) {
				printf("\rWriting new index for frame %d...", nFrame);
			}
			nFrame++;
		}
	}
	fclose(fp2);
	fclose(fp3);

	printf("\n");	
	free(idx_file);free(idxn_file);
	return(0);
}

