#!/usr/bin/perl

#
# vdrrecinfo.pl
#
# Writes a file recinfo.conf containing information about the
# recording, such as start, end, channel, etc.
#
# This code is based on marks2pts of the vdr-sharemarks project
#

use strict;
use warnings;
use LWP::Simple;
use Time::Local; 

my $CHANNELS_CONF = "/etc/vdr/channels.conf";
my $TIMERS_CONF = "/etc/vdr/timers.conf";
my $DEBUG_LOG = "/tmp/vdrrecinfo.log";
my $DEBUG = 1;
my $FAKE_SOURCE = "S19.2E";

##
## no configuration changes should be required below this line
##
create_rec_info($ARGV[0], $ARGV[1]);

sub create_rec_info {
	my $when = shift;
	my $rec_dir = shift;
    dprint ( (join " ", @ARGV) . "\n");
    if (! $when) {
        dprint( "No parameters received, exiting\n");
        exit 1;
    }
    dprint ("vdrrecinfo invoked with parameter $when at " . (localtime()) . "\n");

    if ($when !~ "before") {
        dprint( "Not at start of a recording, but at $when, exiting\n");
        exit;
    }
    if(-e "$rec_dir/recinfo.conf") {
        dprint( "recinfo.conf already exists, won't overwrite\n");
	exit;    
    }
    #my $rec_dir = shift @ARGV;
    
    dprint("looking for timers\n");
    my @timer_hits = @{ get_channel_from_timers($rec_dir) };
    if(!@timer_hits) {
        dprint("could not extract the channelID from timers.conf\n");
        die "could not extract the channelID from timers.conf";
    }
    my $conf_info = $timer_hits[0] . "rstart  ".(time())."\n";	# used by clipinc.pl
    
    if(scalar(@timer_hits) != 1) {
            dprint("Failed to get timer\n");
            die "Failed to determine the channel-id of this recording\n";
    }
    dprint("Writing record info $rec_dir/recinfo.conf $conf_info\n");
    
    
    # Write delayed, probably because VDR creates the $recdir afterwards...
    my $pid;
    if ($pid = fork) {
        dprint("parent exiting\n");
        exit;
        # parent here
    }
    else {
        dprint("child sleeping\n");
        sleep(10);
        open(FH, ">$rec_dir/recinfo.conf") || die "Can not open $rec_dir/recinfo.conf: $!\n";
        binmode(FH);
        print FH $conf_info;
        close FH;
        dprint("Wrote conf\n");
        exit;
    }
}
    
sub get_channel_from_timers {
    my $rec_dir = shift;
    if (!$rec_dir) {
        die "recording dir parameter missing\n";
    }
    my $date = (split /\//, $rec_dir)[-1];
    my $dir_title; # = (split /\//, $rec_dir)[-2];
    
    # Note: in marks2pts, the [.:] didn't escape the dot
    my($year, $mon, $mday, $hour, $min) = $date =~ /(\d{4})-(\d{2})-(\d{2})[\.:](\d{2})[\.:](\d{2})/;
    if((!$year) || (!$mon) || (!$mday) || (!$hour) || (!$min)) {
        die "Could not extract a date from $date\n";
    }
    my $start = tm2unix("$mday.$mon.$year $hour:$min");

    open(FH, $TIMERS_CONF) || die "could not open file: $!\n";
    binmode FH;
    my @timers = (<FH>);
    close FH;   
    
    my @hits;
    foreach (@timers) {
        $dir_title = (split(/\//, $rec_dir))[-2];
        dprint( "Checking $_\n");
        my($status, $cid, $tday, $tstime, $tetime, $prio, $lifetime, $title) = (split(/:/, $_))[0..7];
        my $tshour = substr($tstime, 0, 2);
        my $tsmin  = substr($tstime, 2, 2);
        my $tehour = substr($tetime, 0, 2);
        my $temin  = substr($tetime, 2, 2);
        
        if (!$status) {
            next;
        }
        
        #print "Comparing channel $cid at $tday from $tshour:$tsmin to $tehour:$temin\n";
        
        my $s_uxtime = tm2unix("$mday.$mon.$year  $tshour:$tsmin");
        my $e_uxtime = tm2unix("$mday.$mon.$year  $tehour:$temin");
        if($e_uxtime < $s_uxtime) {
                $e_uxtime += 86400;
        }
        
        if ($tday !~ /^\d{1,2}$/) {
            dprint( "This is a repeating timer, support not well tested\n");
            my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = localtime($s_uxtime);
            
            if(substr($tday, ($wday - 1), 1) !~ /-/) {
                dprint( "ok, this timer should be active today...\n");
                $tday = $mday;
            } else {
                dprint( "not active on $wday: " . substr($tday, $wday, 1) ."\n");
                next;
            }
            $dir_title = (split /\//, $rec_dir)[-3];
        }
        
        dprint( "Title is $title\n");
        if ($title =~ /~/) {
            ($title) = $title =~/.+?~(.+)/;
            dprint( "Found ~ in timer, title is now: $title\n");
        }   
        $title =~ s/[^-a-zA-Z1-9]/.+?/gm;
        #print "title is now $title\n";
        my $title_regex = qr/$title/;

        #print "start is $start and s_uxtime is $s_uxtime to $e_uxtime\n";
       
        if ((abs($start - $s_uxtime) < 180) && ($dir_title =~ $title_regex)) {
            #print "Timerentry confirmed, pushing to hitlist\n";
            my $nchid = channel2nchid($cid);
            if (! $nchid) {die "conversion to nchid did not work\n";}
            push @hits, "channel $cid\nchid    $nchid\nstart   $s_uxtime\nend     $e_uxtime\n";
        } else {
            #print "time mismatch: \n";
        }
    }
    #print @hits;
    return \@hits;
}

sub channel2nchid {
    my $channel = shift;
    if (! -e $CHANNELS_CONF) {
        die "Can not find channel.conf at $CHANNELS_CONF\n";
    }
    
    my ($tsource, $tnid, $tfreq, $tsid, undef) = split /-/, $channel;
    my $ttid = 0;
    #print "got $tsource, $tfreq, $tnid, $tsid from timer channel\n";
    
    open FH, "<$CHANNELS_CONF" || die "Could not open $CHANNELS_CONF: $!\n";
    binmode FH;
    my @channels = (<FH>);
    close FH;
    foreach (@channels) {
        if ($_ =~ /^:/) {next}
        my (undef, $freq, undef, $source, undef, $vpid, $apid, $tpid, undef, $sid, $nid, $tid, undef) =
               split /:/, $_;
        #print "Analysing channel entry $_";
        #print "extracted $freq, $source, $vpid, $apid, $tpid, $sid, $nid, $tid, \n";
        if ($nid && $tid && $tnid && (! $ttid))  {
            #print "Found nid and tid, assuming AutoPID and looking for TID instead of freq\n";
            $ttid = $tfreq;
            $tfreq = 0;
        }
        
        #print "Testing $freq, , $source, $vpid, $apid, $sid\n";
        if ($apid =~/(\d+)[,;]\d+/) {
            $apid = $1;
            dprint( "changing audio PID $apid to $apid\n");
        }
        if (($source eq "C") && (! $ttid)) {
            $freq  = substr ($freq, 0, 3);
            $tfreq = substr ($tfreq, 0, 3);
        }
        #print "Testing from channel: $freq, $source, $vpid, $apid, $sid, $tid against from Timer: $tfreq, $tsource, $tsid,  $ttid\n";
        if (($sid     == $tsid) &&
            ($source  eq $tsource)) {
            #print "matched source and sid\n";
            if (($tid == $ttid) || ($freq == $tfreq)) {
                #print "Found channel: $_\n";
                if ($FAKE_SOURCE) {
                    $source = $FAKE_SOURCE;
                }
            return "$source-$sid-$vpid-$apid";
            }
        }
    }
    return 0;
}

# ------------------------------------------
sub tm2unix {
# ------------------------------------------
    my $value = shift || time;
    my $calc;

    if($value =~ /^\d+$/) {
        my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) =
        localtime($value);
        $calc = sprintf('%02d.%02d.%04d %02d:%02d',
        $mday, $mon+1, $year + 1900,
        $hour, $min);
    } elsif ($value =~ /(\d{2})\.(\d{2})\.(\d{4})\s+(\d{2})\:(\d{2})\:(\d{2})/)
        {
        $calc = timelocal($6, $5, $4, $1, $2-1, $3);
    } elsif($value =~ /(\d{2})\.(\d{2})\.(\d{4})\s+(\d{2})\:(\d{2})/) {
        $calc = timelocal(0, $5, $4, $1, $2-1, $3);
    } else {
        return error('Wrong Format: %s!', $value);
    }
 return $calc;
}

sub dprint {
    if(!$DEBUG) {
        return; # 
    }
    my $msg = shift;
    open DFH, ">>$DEBUG_LOG";
    binmode DFH;
    print $msg;
    print DFH $msg;
    close DFH;
}
